#define TEST_VALUE_1	0xdeadbeef
#define TEST_VALUE_2	0xabcd1234

#define TEST_STRING_1	"hello world"

#ifndef __ASSEMBLY__
extern struct fdt_header _test_tree1;
extern struct fdt_header _truncated_property;
#endif /* ! __ASSEMBLY */
